@@ -14,12 +14,9 @@ from subprocess import Popen, PIPE
 from tempfile import mkstemp
 
 try:
-    from cStringIO import StringIO
+    from StringIO import StringIO
 except ImportError:
-    try:
-        from StringIO import StringIO
-    except ImportError:
-        from io import StringIO
+    from io import StringIO
 
 ROOT_DIR = os.path.split(os.path.abspath(os.path.dirname(__file__)))[0]
 
@@ -161,6 +158,88 @@ def foo():
     def test_supported_fixes(self):
         self.assertIn('E101', [f[0] for f in autopep8.supported_fixes()])
 
+    def test_shorten_comment(self):
+        self.assertEqual('# ' + '=' * 72 + '\n',
+                         autopep8.shorten_comment('# ' + '=' * 100 + '\n',
+                                                  '\n',
+                                                  max_line_length=79))
+
+    def test_shorten_comment_should_not_split_numbers(self):
+        line = '# ' + '0' * 100 + '\n'
+        self.assertEqual(line,
+                         autopep8.shorten_comment(line,
+                                                  newline='\n',
+                                                  max_line_length=79))
+
+    def test_shorten_comment_should_not_split_words(self):
+        line = '# ' + 'a' * 100 + '\n'
+        self.assertEqual(line,
+                         autopep8.shorten_comment(line,
+                                                  newline='\n',
+                                                  max_line_length=79))
+
+    def test_shorten_comment_should_not_modify_special_comments(self):
+        line = '#!/bin/blah ' + ' x' * 90 + '\n'
+        self.assertEqual(line,
+                         autopep8.shorten_comment(line,
+                                                  newline='\n',
+                                                  max_line_length=79))
+
+    def test_format_block_comments(self):
+        self.assertEqual(
+            '# abc',
+            autopep8.format_block_comments('#abc'))
+
+        self.assertEqual(
+            '# abc',
+            autopep8.format_block_comments('####abc'))
+
+    def test_format_block_comments_with_multiple_lines(self):
+        self.assertEqual(
+            """
+# abc
+  # blah blah
+    # four space indentation
+''' #do not modify strings
+#do not modify strings
+#do not modify strings
+#do not modify strings'''
+#
+""".lstrip(),
+            autopep8.format_block_comments("""
+# abc
+  #blah blah
+    #four space indentation
+''' #do not modify strings
+#do not modify strings
+#do not modify strings
+#do not modify strings'''
+#
+""".lstrip()))
+
+    def test_format_block_comments_should_not_corrupt_special_comments(self):
+        self.assertEqual(
+            '#: abc',
+            autopep8.format_block_comments('#: abc'))
+
+    def test_fix_file(self):
+        self.assertIn(
+            'import ',
+            autopep8.fix_file(
+                filename=os.path.join(ROOT_DIR, 'test', 'example.py')))
+
+    def test_normalize_line_endings(self):
+        self.assertEqual(
+            'abc\ndef\n123\nhello\nworld\n',
+            autopep8.normalize_line_endings(
+                'abc\ndef\n123\rhello\r\nworld\r'))
+
+    def test_normalize_line_endings_with_crlf(self):
+        self.assertEqual(
+            'abc\r\ndef\r\n123\r\nhello\r\nworld\r\n',
+            autopep8.normalize_line_endings(
+                'abc\ndef\r\n123\r\nhello\r\nworld\r'))
+
 
 class TestFixPEP8Error(unittest.TestCase):
 
@@ -174,15 +253,9 @@ class TestFixPEP8Error(unittest.TestCase):
         with open(self.tempfile[1], 'w') as temp_file:
             temp_file.write(line)
         opts, _ = autopep8.parse_args([self.tempfile[1]] + list(options))
-        sio = StringIO()
-        if sys.version_info[0] < 3:
-            output = codecs.getwriter(locale.getpreferredencoding())(sio)
-        else:
-            output = sio
-        autopep8.fix_file(filename=self.tempfile[1],
-                          opts=opts,
-                          output=output)
-        self.result = output.getvalue()
+        self.result = autopep8.fix_file(
+            filename=self.tempfile[1],
+            opts=opts)
 
     def test_e101(self):
         line = """
@@ -328,8 +401,7 @@ try:
                   ['-vvv', '--select=E101', '--diff',
                    os.path.join(ROOT_DIR, 'test', 'e101_example.py')],
                   stdout=PIPE, stderr=PIPE)
-        output, error = [x.decode('utf-8') for x in p.communicate()]
-        self.assertIn('Not fixing E101 on line', error)
+        output = [x.decode('utf-8') for x in p.communicate()][0]
         self.assertEqual('', output)
 
     def test_e111_short(self):
@@ -751,6 +823,24 @@ def draw(self):
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e127_with_backslash(self):
+        line = r"""
+if True:
+    if True:
+        self.date = meta.session.query(schedule.Appointment)\
+            .filter(schedule.Appointment.id ==
+                                      appointment_id).one().agenda.endtime
+""".lstrip()
+        fixed = r"""
+if True:
+    if True:
+        self.date = meta.session.query(schedule.Appointment)\
+            .filter(schedule.Appointment.id ==
+                    appointment_id).one().agenda.endtime
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e12_with_backslash(self):
         line = r"""
 if True:
@@ -1087,8 +1177,8 @@ class Foo(object):
         self.assertEqual(self.result, fixed)
 
     def test_e304_with_comment(self):
-        line = "@contextmanager\n#comment\n\ndef f():\n    print 1\n"
-        fixed = "@contextmanager\n#comment\ndef f():\n    print 1\n"
+        line = "@contextmanager\n# comment\n\ndef f():\n    print 1\n"
+        fixed = "@contextmanager\n# comment\ndef f():\n    print 1\n"
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
@@ -1327,8 +1417,55 @@ def dummy():
         self._inner_setup(line)
         self.assertEqual(self.result, line)
 
+    def test_e501_should_not_break_on_dot(self):
+        line = """
+if True:
+    if True:
+        raise xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx('xxxxxxxxxxxxxxxxx "{d}" xxxxxxxxxxxxxx'.format(d='xxxxxxxxxxxxxxx'))
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, line)
+
+    def test_e501_with_comment(self):
+        line = """123
+                        # This is a long comment that should be wrapped. I will wrap it using textwrap to be within 72 characters.
+"""
+        fixed = """123
+                        # This is a long comment that should be wrapped. I will
+                        # wrap it using textwrap to be within 72 characters.
+"""
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e501_with_comment_should_not_modify_docstring(self):
+        line = '''
+def foo():
+    """
+                        # This is a long comment that should be wrapped. I will wrap it using textwrap to be within 72 characters.
+    """
+'''.lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, line)
+
+    def test_e501_should_only_modify_last_comment(self):
+        line = """123
+                        # This is a long comment that should be wrapped. I will wrap it using textwrap to be within 72 characters.
+                        # 1. This is a long comment that should be wrapped. I will wrap it using textwrap to be within 72 characters.
+                        # 2. This is a long comment that should be wrapped. I will wrap it using textwrap to be within 72 characters.
+                        # 3. This is a long comment that should be wrapped. I will wrap it using textwrap to be within 72 characters.
+"""
+        fixed = """123
+                        # This is a long comment that should be wrapped. I will wrap it using textwrap to be within 72 characters.
+                        # 1. This is a long comment that should be wrapped. I will wrap it using textwrap to be within 72 characters.
+                        # 2. This is a long comment that should be wrapped. I will wrap it using textwrap to be within 72 characters.
+                        # 3. This is a long comment that should be wrapped. I
+                        # will wrap it using textwrap to be within 72
+                        # characters.
+"""
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e501_should_not_interfere_with_non_comment(self):
-        """This is why we've disabled comment shortening."""
         line = '''
 
 """
@@ -1338,12 +1475,29 @@ def dummy():
         self._inner_setup(line)
         self.assertEqual(self.result, line)
 
-    def test_e501_should_not_break_on_dot(self):
-        line = """
-if True:
-    if True:
-        raise xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx('xxxxxxxxxxxxxxxxx "{d}" xxxxxxxxxxxxxx'.format(d='xxxxxxxxxxxxxxx'))
+    def test_e501_should_cut_comment_pattern(self):
+        line = """123
+# -- Useless lines ----------------------------------------------------------------------
+321
+"""
+        fixed = """123
+# -- Useless lines -------------------------------------------------------
+321
+"""
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e501_with_function_should_not_break_on_colon(self):
+        line = r"""
+class Useless(object):
+    def _table_field_is_plain_widget(self, widget):
+        if widget.__class__ == Widget or\
+                (widget.__class__ == WidgetMeta and Widget in widget.__bases__):
+            return True
+
+        return False
 """.lstrip()
+
         self._inner_setup(line)
         self.assertEqual(self.result, line)
 
@@ -1408,7 +1562,7 @@ time.strftime('%d-%m-%Y')
             line = unicode(line, 'utf-8')
             fixed = unicode(fixed, 'utf-8')
             self._inner_setup(line.encode('utf-8'))
-            self.assertEqual(self.result, fixed.encode('utf-8'))
+            self.assertEqual(self.result, fixed)
         else:
             self._inner_setup(line)
             self.assertEqual(self.result, fixed)
@@ -1500,27 +1654,37 @@ raise IOError('abc '
     def test_e711(self):
         line = 'foo == None\n'
         fixed = 'foo is None\n'
-        self._inner_setup(line)
+        self._inner_setup(line, options=['--aggressive'])
         self.assertEqual(self.result, fixed)
 
     def test_e711_in_conditional(self):
         line = 'if foo == None and None == foo:\npass\n'
         fixed = 'if foo is None and None == foo:\npass\n'
-        self._inner_setup(line)
+        self._inner_setup(line, options=['--aggressive'])
         self.assertEqual(self.result, fixed)
 
     def test_e711_in_conditional_with_multiple_instances(self):
         line = 'if foo == None and bar == None:\npass\n'
         fixed = 'if foo is None and bar is None:\npass\n'
-        self._inner_setup(line)
+        self._inner_setup(line, options=['--aggressive'])
         self.assertEqual(self.result, fixed)
 
     def test_e711_with_not_equals_none(self):
         line = 'foo != None\n'
         fixed = 'foo is not None\n'
-        self._inner_setup(line)
+        self._inner_setup(line, options=['--aggressive'])
         self.assertEqual(self.result, fixed)
 
+    def test_e711_should_not_modify_sql_alchemy_query(self):
+        line = 'filter(User.name == None)\n'
+        self._inner_setup(line, options=['--aggressive'])
+        self.assertEqual(self.result, line)
+
+    def test_e711_should_not_modify_sql_alchemy_query_with_not_equals(self):
+        line = 'filter(User.name != None)\n'
+        self._inner_setup(line, options=['--aggressive'])
+        self.assertEqual(self.result, line)
+
     def test_e721(self):
         line = "type('') == type('')\n"
         fixed = "isinstance('', type(''))\n"
@@ -1552,15 +1716,10 @@ class TestFixPEP8Warning(unittest.TestCase):
         with open(self.tempfile[1], 'w') as temp_file:
             temp_file.write(line)
         opts, _ = autopep8.parse_args([self.tempfile[1]] + list(options))
-        sio = StringIO()
-        if sys.version_info[0] < 3:
-            output = codecs.getwriter(locale.getpreferredencoding())(sio)
-        else:
-            output = sio
-        autopep8.fix_file(filename=self.tempfile[1],
-                          opts=opts,
-                          output=output)
-        self.result = output.getvalue()
+
+        self.result = autopep8.fix_file(
+            filename=self.tempfile[1],
+            opts=opts)
 
     def test_w191_should_ignore_multiline_strings(self):
         line = """
@@ -1718,7 +1877,8 @@ a.has_key(
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
-    @unittest.skipIf(sys.version_info[0] < 3, 'needs to be fixed on Python 2')
+    @unittest.skipIf(sys.version_info < (2, 6, 4),
+                     'older versions of 2.6 may be buggy')
     def test_w601_with_non_ascii(self):
         line = """
 # -*- coding: utf-8 -*-
@@ -1736,7 +1896,7 @@ correct = 'good syntax ?' in dict()
             line = unicode(line, 'utf-8')
             fixed = unicode(fixed, 'utf-8')
             self._inner_setup(line.encode('utf-8'))
-            self.assertEqual(self.result, fixed.encode('utf-8'))
+            self.assertEqual(self.result, fixed)
         else:
             self._inner_setup(line)
             self.assertEqual(self.result, fixed)
@@ -1841,12 +2001,11 @@ correct = 'good syntax ?' in dict()
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
-    # Doesn't work because pep8 doesn't seem to work with CR line endings
-    #def test_w602_escaped_cr(self):
-    #    line = 'raise ValueError, \\\r"hello"\n'
-    #    fixed = 'raise ValueError("hello")\n'
-    #    self._inner_setup(line)
-    #    self.assertEqual(self.result, fixed)
+    def test_w602_escaped_cr(self):
+        line = 'raise ValueError, \\\r"hello"\n'
+        fixed = 'raise ValueError("hello")\n'
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
 
     def test_w602_multiple_statements(self):
         line = 'raise ValueError, "hello";print 1\n'
@@ -2055,20 +2214,17 @@ class TestSpawnPEP8Process(unittest.TestCase):
         with open(self.tempfile[1], 'w') as temp_file:
             temp_file.write(line)
         opts, _ = autopep8.parse_args(list(options) + [self.tempfile[1]])
-        sio = StringIO()
 
         # Monkey patch pep8 to trigger spawning
         original_pep8 = autopep8.pep8
         try:
             autopep8.pep8 = None
-            autopep8.fix_file(filename=self.tempfile[1],
-                              opts=opts,
-                              output=sio)
+            self.result = autopep8.fix_file(
+                filename=self.tempfile[1],
+                opts=opts)
         finally:
             autopep8.pep8 = original_pep8
 
-        self.result = sio.getvalue()
-
     def test_basic(self):
         line = "print('abc' )    \n1*1\n"
         fixed = "print('abc')\n1 * 1\n"
@@ -2090,23 +2246,44 @@ class TestSpawnPEP8Process(unittest.TestCase):
         self._inner_setup(line, options=['--max-line-length=40'])
         self.assertEqual(self.result, fixed)
 
+    def test_format_block_comments(self):
+        line = """
+foo(  )
+# abc
+bar()#bizz
+  #blah blah
+    #four space indentation
+if True:
+    1
+""".lstrip()
+        fixed = """
+foo()
+# abc
+bar()  # bizz
+  # blah blah
+    # four space indentation
+if True:
+    1
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+
+@contextlib.contextmanager
+def temporary_file_context(text):
+    tempfile = mkstemp()
+    with open(tempfile[1], 'w') as temp_file:
+        temp_file.write(text)
+    yield tempfile[1]
+    os.remove(tempfile[1])
+
 
 class TestCoverage(unittest.TestCase):
 
-    def setUp(self):
-        self.tempfile = mkstemp()
-
-    def tearDown(self):
-        os.remove(self.tempfile[1])
-
-    def _inner_setup(self, line):
-        with open(self.tempfile[1], 'w') as temp_file:
-            temp_file.write(line)
-
     def test_fixpep8_class_constructor(self):
         line = "print 1\nprint 2\n"
-        self._inner_setup(line)
-        pep8obj = autopep8.FixPEP8(self.tempfile[1], None)
+        with temporary_file_context(line) as filename:
+            pep8obj = autopep8.FixPEP8(filename, None)
         self.assertEqual("".join(pep8obj.source), line)
 
     def test_no_argument(self):
@@ -2125,6 +2302,26 @@ class TestCoverage(unittest.TestCase):
             except SystemExit as e:
                 self.assertEqual(e.code, 2)
 
+    def test_standard_out_should_use_native_line_ending(self):
+        line = '1\r\n2\r\n3\r\n'
+        with temporary_file_context(line) as filename:
+            process = Popen(list(AUTOPEP8_CMD_TUPLE) +
+                            [filename],
+                            stdout=PIPE)
+            self.assertEqual(
+                os.linesep.join(['1', '2', '3', '']),
+                process.communicate()[0].decode('utf-8'))
+
+    def test_standard_out_should_use_native_line_ending_with_cr_input(self):
+        line = '1\r2\r3\r'
+        with temporary_file_context(line) as filename:
+            process = Popen(list(AUTOPEP8_CMD_TUPLE) +
+                            [filename],
+                            stdout=PIPE)
+            self.assertEqual(
+                os.linesep.join(['1', '2', '3', '']),
+                process.communicate()[0].decode('utf-8'))
+
 
 @contextlib.contextmanager
 def disable_stderr():
